
Require Import Setoid.
Require Import Coq.Classes.Morphisms.



(* suppose we want to do a proof that 
   for some function from the naturals into the naturals

   F2eX: f(f(f(x))) = x,
   F3eX: f(f(x)) = x 
   ⊢ f(x) = x


   the first insinct is to just try rewriting it,

   but if you rewrite the orignal goal by the second hypothosis you get
  
  rewrite <- F2eX.

  F2eX: f(f x) = x,
  F3eX: f(f(f x)) = x 
    ⊢ f(f(f(x))) = f(f(x))

  the lhs now lines up with a proof we have, but in the process we altered the rhs

  the fix for this is to say that we want to rewrite only the first occurence of x,

  rewrite <- F2eX at 1.
  
  F2eX: f(f(x)) = x,
  F3eX: f(f(f(x))) = x 
    ⊢ f(f(f(x))) = x
  
  clearly this goal is solveable,

  a clever user might rewrite in the fist hypothosis

  rewrite F2eX in F2eX.

  F2eX: f(f(x)) = x,
  F3eX: f(x) = x 
    ⊢ f(x) = x

  but this still requires some ammount of cleverness. Fortunatly coq provides a hamemr for this situation: `congruence`.


  invoking `congruence.` simply solves this goal.


  But what if you where not dealing with normal equality. Many types in coq have a sense of equaltiy that is not captured by Logic.eq. For example consider the rationals. Coq represents these as just a pair of a signed numerator and a positive denominator, it then defines Qeq (a#b) (c#d) := (a * d == b * c).

  So what if we wanted to prove
  F2eX: Qeq (f(f(x))) x,
  F3eX: Qeq (f(f(f(x)))) x 
    ⊢ Qeq (f(x)) x

  well the first thing to note is that this is not automaticaly true for all f, consider the function that maps:
  f(1#1) -> f(1#2)
  f(1#2) -> f(2#2)
  f(2#2) -> f(2#2)
  and maps all other values to 0#1.


  clearly this satifies that
  f(f(1#1)) == 1#1, and f(f(f(1#1))) == 1#1, but not that f(1#1) == 1#1.

  This is because f is not repectfull of `==`, namely that even though 1#1 == 2#2, it is not the case (f(1#1) == 2#2). Coq lets us keep track of this using `Proper` instances, the specific constraint we want is `Proper (Qeq ==> Qeq) f`. So with that asside, let's say we do have that. Then
  fProper: Proper (Qeq ==> Qeq) f
  F2eX: Qeq (f(f(f(x)))) x,
  F3eX: Qeq (f(f(x))) x 
    ⊢ Qeq (f(f(f(x)))) x

  can be solved by 

  `rewrite F2eX in F3eX; exact F3eX`.

  but congruence doesn't work.

  In this MQP I aimed to create a version of congruence that supports other equivlence relations
*)




Lemma simpleRelation:
  forall (f: nat -> nat) (x: nat), f (f x) = x -> f (f (f x)) = x -> f x = x.
Proof.
  intros f x F2eX F3eX.
  rewrite <- F2eX at 1.
  exact F3eX.
Qed.

Lemma simpleRelationAlt:
  forall (f: nat -> nat) (x: nat), f (f x) = x -> f (f (f x)) = x -> f x = x.
Proof.
  intros f x F2eX F3eX.
  rewrite F2eX in F3eX.
  exact F3eX.
Qed.



Lemma simpleRelationWithCongruence:
  forall (f: nat -> nat) (x: nat), f (f x) = x -> f (f (f x)) = x -> f x = x.
Proof.
  intros f x F2eX F3eX.
  congruence.
Qed.




Require Import QArith.
Local Open Scope Q_scope.



Lemma simpleRelationOverQ:
  forall (f: Q -> Q)  (x: Q), (Proper (Qeq ==> Qeq) f) -> f (f x) == x -> f (f (f x)) == x -> f x == x.
Proof.
  intros f x fproper F2eX F3eX.
  rewrite F2eX in F3eX.
  easy.
Qed.



Lemma harder:
  forall (f: nat -> nat) (x: nat), 
    f (f (f x)) = x -> 
    f (f (f (f (f x)))) = x ->
    f x = x.
Proof.
  intros f x F3eX F5eX.
  rewrite <- F5eX at 1.
  rewrite F3eX.
  apply F3eX.
Qed.

Lemma harder2:
  forall (f: nat -> nat) (x: nat), 
  f (f (f x)) = x -> 
  f (f (f (f (f x)))) = x ->
  f x = x.
Proof.
  intros f x F3eX F5eX.
  rewrite F3eX in F5eX.
  rewrite F5eX in F3eX.
  exact F3eX.
Qed.

