Set Universe Polymorphism.


Require Import Ltac2.Ltac2.

Module UfProof.
  Ltac2 Type rec 'a t
    := [ Trans (int, 'a t, int, 'a t, int)
        | Sym( int, int, 'a t )
        | Cong (int, int,
                int, int,
                'a t, 'a t)
        | Ax (int, int, 'a ) 
        | Refl ( int ) ].
  
  Ltac2 sym (oldLhs: int) (oldRhs: int) (x: 'a t)
    := match x with
       | Sym _ _ y => y
       | Refl r => Refl r
       | _ => Sym oldLhs oldRhs  x
       end.
  
  (* note the argument order is funny for readablility *)
  Ltac2 trans (a: int) (x: 'a t) (b: int) (y: 'a t) (c: int)
    := match x with
       | Refl _ => y
       | _ => match y with
              | Refl _ => x
              | _ => Trans a x b y c
              end
        end.




  Ltac2 refl (x: int) := Refl x.
  Ltac2 cong (f: int) (f' : int) (x : int) (x' : int) (fe: 'a t) (xe: 'a t) := Cong f f' x x' fe xe.
  Ltac2 axiom (lhs: int) (rhs: int) (pf: 'a) := Ax lhs rhs pf.

  (* acts like:
      ∀ x x' 
        y y', x = x' 
                -> y = y'
                -> x = y -> x' = y'  *)

  Ltac2 transportEq (x: int) (x': int) (y: int) (y': int) (h1: 'a t) (h2: 'a t) (h3: 'a t) := trans x' (trans x' (sym x x' h1) x h3 y) y h2 y'.
End UfProof.



(* Module IntSet.
  Ltac2 Type t := int list.
  Ltac2 rec union (a : t) (b: t)
    := match a with
       | [] => b
       | a0 :: ar => 
          match b with
          | [] => a
          | b0 :: br => 
              match (Int.lt a0 b0) with
              | true => a0 :: union ar b
              | false => 
                  match (Int.equal a0 b0) with
                  | true => a0 :: union ar br
                  | false => b0 :: union a br
                  end
              end
          end
       end.
End IntSet. *)



(* union find that tracks classes *)
Module UF.
  Ltac2 Type EClass 
    := {
      size : int ;
      leftPred : int list;
      rightPred : int list;
    }.

  Ltac2 Type 'a Node 
    := [ Repr ( EClass ) (* class  *)
       | Node ( 'a UfProof.t, int ) (* proof, parent *)
      ].

  Ltac2 Type 'a t := ('a Node) array.
  
  (* finds class, returns tuple of ( repTerm, eclass, proof  )  *)
  (* side effects: optimizes structure of the UF tree  *)
  Ltac2 rec findClass ( uf : 'a t ) ( term: int )
    := match (Array.get uf term) with
      | Repr eclass => ( term, eclass, UfProof.Refl term )
      | Node proof parent =>
            let ( repTerm, eclass, prf) := findClass uf parent in
            let newProof := UfProof.trans term proof parent prf repTerm in
              Array.set uf term (Node newProof repTerm) ;
              (repTerm, eclass,  newProof)
      end.
  
  

  
  (*
    unfies two terms, returing the *origional* left and right parents of each 
  *)
  Ltac2 union (uf : 'a t ) (lhs: int) (rhs: int) (proof: 'a UfProof.t)
    := let (lhsRep, lhsClass, lhsPrf) := findClass uf lhs in 
       let (rhsRep, rhsClass, rhsPrf) := findClass uf rhs in 
       match (Int.equal lhsRep rhsRep) with
       | false =>
                let classEqProof := (UfProof.transportEq lhs lhsRep rhs rhsRep lhsPrf rhsPrf proof) in
                match (Int.ge (lhsClass.(size)) (rhsClass.(size))) with
                | true => 
                    (* lhs is larger so it should be the new root *)
                    Array.set uf lhsRep (Repr {
                      size := Int.add (lhsClass.(size)) (rhsClass.(size)) ;
                      leftPred := List.append (lhsClass.(leftPred)) (rhsClass.(leftPred)) ;
                      rightPred := List.append (lhsClass.(rightPred)) (rhsClass.(rightPred))
                    }) ; 
                    Array.set uf rhsRep (Node 
                      (* make a proof that rhsRep = lhsRep *)
                      (UfProof.transportEq rhs rhsRep  lhs lhsRep  rhsPrf lhsPrf (UfProof.sym lhs rhs proof))
                      lhsRep)
                | false =>
                    Array.set uf lhsRep (Node 
                              (* make a proof that rhsRep = lhsRep *)
                              (UfProof.transportEq lhs lhsRep rhs rhsRep lhsPrf rhsPrf proof)
                              rhsRep);
                    Array.set uf rhsRep (Repr {
                      size := Int.add (lhsClass.(size)) (rhsClass.(size)) ;
                      leftPred := List.append (lhsClass.(leftPred)) (rhsClass.(leftPred)) ;
                      rightPred := List.append (lhsClass.(rightPred)) (rhsClass.(rightPred))
                    })
                end;
                Some (
                  (lhsClass.(leftPred), rhsClass.(leftPred)),
                  (lhsClass.(rightPred), rhsClass.(rightPred))
                )
        | true => None
        end.

  Ltac2 isCongruent (uf : 'a t ) (lhs: int) (rhs: int)
    := let (lhsRep, lhsClass, lhsPrf) := findClass uf lhs in 
       let (rhsRep, rhsClass, rhsPrf) := findClass uf rhs in
        Int.equal lhsRep rhsRep.

  Ltac2 congruent (uf : 'a t ) (lhs: int) (rhs: int)
    := let (lhsRep, lhsClass, lhsPrf) := findClass uf lhs in 
       let (rhsRep, rhsClass, rhsPrf) := findClass uf rhs in 
       match (Int.equal lhsRep rhsRep) with
       | false => None
       | true => Some (UfProof.trans lhs lhsPrf lhsRep (UfProof.sym rhs lhsRep rhsPrf) rhs)
       end.
  Ltac2 fromPreds (x : (int list * int list ) array)
    := Array.map (fun (l, r) => Repr ({ size := 1 ; leftPred := l ; rightPred := r })) x.
End UF.




Module Helpers.
  Ltac2 rec iterDown (c: int) (f: int -> unit)
    := match (Int.equal c 0) with
       | true => ()
       | false => let next := Int.sub c 1 in f next; iterDown next f
       end.
  Ltac2 update arr index f := Array.set arr index (f (Array.get arr index)).
End Helpers.


Module CC.
  Ltac2 Type LookupEntry := [ Atom | Children(int, int) ].
  Ltac2 Type 'a t
    := {
        termLookup : LookupEntry array ;
        unionFind : 'a UF.t ;
        mutable equationsQue : (int * int * 'a UfProof.t) list ;
        mutable congruencesQue : (int list * int list) list
      }.

  Ltac2 doEquation
      (cc: 'a t)
      (lhs: int)
      (rhs: int)
      (proof: 'a UfProof.t)
    := match (UF.union (cc.(unionFind)) lhs rhs proof) with
       | None => ()
       | Some x => let (lp, rp) := x in
          cc.(congruencesQue) := lp :: rp :: cc.(congruencesQue)
       end.
  
  Ltac2 doCongruence
    (cc: 'a t)
    (lhs: int)
    (rhs: int)
    := match (Array.get (cc.(termLookup)) lhs) with | Atom => ()
       | Children lf lx => 
       match (Array.get (cc.(termLookup)) rhs) with | Atom => ()
       | Children rf rx => 
       match (UF.isCongruent (cc.(unionFind)) lhs rhs) with | true => ()
       | false =>
       match (UF.congruent (cc.(unionFind)) lf rf) with | None => ()
       | Some fProof => 
       match (UF.congruent (cc.(unionFind)) lx rx) with | None => ()
       | Some xProof => 
        doEquation cc lhs rhs (UfProof.cong lf rf lx rx fProof xProof)
       end
       end
       end
       end
       end.

  Ltac2 step (cc: 'a t)
    := match (cc.(congruencesQue)) with
       | congr :: congrQue =>
          cc.(congruencesQue) := congrQue;
          let (lhss, rhss) := congr in
          List.iter (fun lhs => List.iter (doCongruence cc lhs) rhss) lhss;
          true
       | [] =>
          match (cc.(equationsQue)) with
          | e :: eqQue => 
            cc.(equationsQue) := eqQue;
            let (lhs, rhs, prf) := e in
            doEquation cc lhs rhs prf;
            true
          | [] => false
          end
       end.
  Ltac2 rec run (cc: 'a t)
    := match (step cc) with
       | true => run cc
       | false => ()
       end.

  Ltac2 init (a : LookupEntry array) (eqns: (int * int * 'a UfProof.t) list  )
    := let predLookups := Array.make (Array.length a) ( [], [] ) in
        Helpers.iterDown (Array.length a)
          (fun i => match (Array.get a i) with
                    | Atom => ()
                    | Children lc rc =>
                          Helpers.update predLookups lc (fun (l, r) => (i :: l, r));
                          Helpers.update predLookups rc (fun (l, r) => (l,  i :: r))
                    end);
        {
          termLookup := a ;
          unionFind := UF.fromPreds predLookups ;
          congruencesQue := [] ;
          equationsQue := eqns
        }.
  Ltac2 congruent (cc : 'a t ) (lhs: int) (rhs: int) := UF.congruent (cc.(unionFind)) lhs rhs.
End CC.



Module Testing.
  (* See example 4.4.4 in Term Rewriting and All That *)
  Ltac2 terms (u: unit) := Array.of_list [ 
        (* 0 = f *) CC.Atom ; 
        (* 1 = x *) CC.Atom ;
        (* 2 = f x *) CC.Children 0 1 ;
        (* 3 = f (f x) *) CC.Children 0 2 ;
        (* 4 = f (f (f x)) *) CC.Children 0 3 
      ].

  Ltac2 equations (u: unit)
    := [ (1,4, UfProof.axiom 1 4 "f^3 (x) = x");
         (1,3, UfProof.axiom 1 3 "f^2 (x) = x") ].

  Ltac2 demo1 (u: unit) := let x := (CC.init (terms ()) (equations ())) in CC.run x;  x.

  (* Ltac2 Eval (demo1 ()). *)
End Testing.

(* helper to build terms that CC likes *)
(* this is janky, and is the unavoidably slow part *)
Module CCBuilder.
  
  Ltac2 Type ('k, 'a) t := {
    (* *reversed* list of term names *)
    mutable termNames : ('k * CC.LookupEntry ) list ;
    mutable equations : (int * int * 'a UfProof.t) list 
  }.

  Ltac2 rec aux_search f lst
    := match lst with
       | [] => None
       | x :: xr => match (f x) with
                    | true => Some (List.length xr)
                    | false => aux_search f xr
                    end
       end.

  Ltac2 makeTerm (builder: ('k,'a) t) (k: 'k) (data: (int * int) option)
    := let old := builder.(termNames) in
       builder.(termNames) := (k, match data with
                                  | Some x => let (l,r) := x in CC.Children l r
                                  | None => CC.Atom
                                  end) :: old; 
       List.length old.

  Ltac2 lazyMakeTerm (builder: ('k,'a) t) (query: 'k -> bool) ( fb : unit -> int )
    := match (aux_search (fun (k,_) => query k) (builder.(termNames))) with
       | Some i => i
       | None => fb ()
       end.

  Ltac2 addEquation builder lhs rhs prf := builder.(equations) := (lhs,rhs,UfProof.axiom lhs rhs prf) :: builder.(equations).


  Ltac2 build (builder : ('k, 'a) t)
    := (
        CC.init
          (Array.of_list (List.rev_map (fun (_,r) => r) (builder.(termNames))))
          (builder.(equations)),
        (Array.of_list (List.rev_map (fun (l,_) => l) (builder.(termNames))))).
  Ltac2 new (u: unit) := { termNames := [] ; equations := [] }.
End CCBuilder.



Module BasicConstr.

  Lemma f_equal_split: forall {A B: Type} (f f': A -> B) (x x': A), f = f' -> x = x' -> f x = f' x'.
  Proof.
    intros.
    now subst.
  Defined.

  Ltac2 rec addTerm (builder : (constr, 'a) CCBuilder.t) (tm : constr)
    :=  CCBuilder.lazyMakeTerm builder (Constr.equal tm) 
          (fun () =>
              lazy_match! tm with
              | ?f ?x => lazy_match! (Constr.type f) with
                         | (_ -> _) => CCBuilder.makeTerm builder tm (Some (addTerm builder f, addTerm builder x))
                         | _ => CCBuilder.makeTerm  builder tm None
                         end
              | _ => CCBuilder.makeTerm  builder tm None
              end).
  Ltac2 rec addEquations (builder : (constr, constr) CCBuilder.t)
    := List.iter (fun (id,_,ty) => 
                      lazy_match! ty with
                      | ?a = ?b => CCBuilder.addEquation builder (addTerm builder a) (addTerm builder b) (Control.hyp id) 
                      | _ => ()
                      end) 
       (Control.hyps ()).

  Ltac2 rec fromProof (lookup: constr array) (prf: constr UfProof.t) 
    := match prf with
       | UfProof.Trans x h y k z 
          => let xt := Array.get lookup x in
             let yt := Array.get lookup y in
             let zt := Array.get lookup z in
             let l := fromProof lookup h in
             let r := fromProof lookup k in
             constr:(@eq_trans _ $xt $yt $zt $l $r)
       | UfProof.Sym x y h 
          => let xt := Array.get lookup x in
             let yt := Array.get lookup y in
             let l := fromProof lookup h in
             constr:(@eq_sym _ $xt $yt $l)
       | UfProof.Cong f f' x x' h k
          => let ft := Array.get lookup f in
             let ft' := Array.get lookup f' in
             let xt := Array.get lookup x in
             let xt' := Array.get lookup x' in
             let l := fromProof lookup h in
             let r := fromProof lookup k in
             constr:(@f_equal_split _ _ $ft $ft' $xt $xt' $l $r)
       | UfProof.Ax l r a => a
       | UfProof.Refl x 
          => let xt := Array.get lookup x in
             constr:(@eq_refl _ $xt)
       end.
  
  Ltac2 rec solveGoal (u: unit)
    := lazy_match! goal with
       | [ |- ?a = ?b ] => 
                            let bld := CCBuilder.new () in
                            addEquations bld;
                            let lhs := addTerm bld a in
                            let rhs := addTerm bld b in
                            let (cc, lookupMap) := CCBuilder.build bld in
                            CC.run cc;
                            match (CC.congruent cc lhs rhs) with
                            | Some prf => let x := fromProof lookupMap prf in Control.refine (fun () => x)
                            | _ => Control.zero (Tactic_failure None)
                            end
        end.
  

  Module Examples.
    Lemma example_4_4_4: forall (f : nat -> nat) (x: nat), f (f (f x)) = x -> f (f x) = x -> f x = x.
    Proof.
      intros.
      solveGoal ().
    Qed.

    Lemma exersize_from_later: forall (f : nat -> nat) (x: nat), f (f (f x)) = x -> f (f (f (f (f x)))) = x -> f (x) = x.
    Proof.
      intros.
      solveGoal ().
    Qed.
    
    Lemma itWorks:
      forall (o: nat -> nat -> nat)
             (x y: nat),
             o x y = o y x ->
             o x x = x ->
             o y y = y ->
             (o (o x x) (o y y)) = o (o x y) (o y x)
             <-> 
             (o (o x y) (o x y) = o x y).
    Proof.
      intros.
      split.
      + intros. solveGoal ().
      + intros. solveGoal ().
    Qed.
  End Examples.
End BasicConstr.




